// This is the main driver file.
// 
// Codon Optimizer - a reference implementation of the CAI optimization 
// and motif engineering algorithm by Anne Condon and Chris Thachuk. 
//
// Copyright (C) 2012 Chris Thachuk
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


// standard libraries
#include <fstream>
#include <iostream>
#include <string>
#include <vector>

// boost libraries
#include <boost/algorithm/string.hpp>
#include <boost/program_options.hpp>
#include <boost/timer/timer.hpp>
namespace po = boost::program_options;

// useful classes from the oligo design library
#include "cai_optimizer.h"
#include "codon_table.h"
#include "dna_sequence.h"
#include "util.h"
namespace od = oligodesign;


// a simple struct to hold command line options passed by user
struct prog_options {
  int start_index;
  int end_index;
  std::string fasta_file;
  std::string forbidden_filename;
  std::string desired_filename;
  std::string opt_seq_filename;
  std::string trace_filename;
  std::vector<od::DNASequence> forbidden;
  std::vector<od::DNASequence> desired;
};

bool init(int, char *[], po::variables_map &, prog_options &options);


int main(int argc, char* argv[])
{
  po::variables_map vm;
  prog_options options;

  if (!init(argc, argv, vm, options)) {
    return 1;
  }

  od::CodonTable ctable;
  // od::codon_optimizer coptimizer;

  std::vector<od::DNASequence>::const_iterator itr;
  std::cout << "Eliminating the following forbidden motifs:\n\n";
  for (itr = options.forbidden.begin(); itr != options.forbidden.end(); ++itr) {
    std::cout << *itr << std::endl;
  }
  std::cout << std::endl;
  
  std::cout << "Adding the following desirable motifs:\n\n";
  for (itr = options.desired.begin(); itr != options.desired.end(); ++itr) {
    std::cout << *itr << std::endl;
  }
  std::cout << std::endl;

  std::ofstream trace(options.trace_filename.c_str());
  std::ofstream opt_file(options.opt_seq_filename.c_str());
  bool warnings_occurred = false;

  try {
    int count = options.end_index - options.start_index + 1;
    std::cout << "Optimizing " << count << " sequence(s)." << std::endl;
    std::cout << std::endl;
    std::cout << "0 %";
    std::cout.flush();
    int progress = 0;
    
    for (int i = options.start_index; i <= options.end_index; i++) {

      std::string sequence;
      std::string sequence_name;
      if (!od::SequenceFromFasta(options.fasta_file.c_str(), sequence, i, sequence_name)) {
        std::cerr << i << " is an invalid index number for a sequence in the FASTA file ";
        std::cerr << options.fasta_file << std::endl;
        std::cerr << "Aborting!" << std::endl;
        throw "Invalid FASTA index";
      }
      
      if(sequence.length() % 3 > 0 ) {
        int new_len = sequence.length() - (sequence.length() % 3);
        sequence = sequence.substr(0,new_len);
        warnings_occurred = true;
        trace << "# Warning: sequence " << i 
              << " length is not a multiple of 3.  It has been truncated." << std::endl;
      }

      od::DNASequence dna(sequence);
      if (!dna.ValidSeq()) {
        warnings_occurred = true;
        trace << "# Warning: sequence " << i
              << " contained characters other than A,C,G or T.  They were replaced as 'C'." << std::endl;
      }

      double score_before = od::CalcCAI(dna, ctable);
      int forbidden_before = od::CountMotifOccurences(dna, options.forbidden);
      int desired_before = od::CountMotifOccurences(dna, options.desired);
 
      boost::timer::cpu_timer opt_timer;
      double score_after = od::OptimizeCAI(dna, ctable, options.forbidden, options.desired);
      opt_timer.stop();
      int forbidden_after = od::CountMotifOccurences(dna, options.forbidden);
      int desired_after = od::CountMotifOccurences(dna, options.desired);
	
      opt_file << sequence_name << std::endl;
      opt_file << dna << std::endl;
      
      trace << i << "\t" << sequence.length() << "\t";
      trace << score_before << "\t" << forbidden_before << "\t" << desired_before << "\t";
      trace << score_after << "\t" << forbidden_after << "\t" << desired_after;
      trace << "\t" << opt_timer.format(6, "%t") << "\n";

      ++progress;
      std::cout << "\r" << (100 * progress) / count << " %";
      std::cout.flush();
    }
  }
  catch (...) { }

  trace.close();
  opt_file.close();

  if (warnings_occurred) {
    std::cout << std::endl;
    std::cout << "Warnings occurred while optimizing sequences.  See '"
              << options.trace_filename << "' for details." << std::endl;
  }

  return 0;
}

bool init(int argc, char *argv[], po::variables_map &vm, prog_options &o) {
  using namespace std;
  using namespace oligodesign;

  try {
    // Declare a group of options that will be
    // allowed only on command line
    po::options_description generic("Generic");
    generic.add_options()
      ("help,h", "produce this help message")
      ;

    // Declare a group of options that will be
    // allowed both on command line and in
    // config file
    po::options_description design_config("Design Specifications");
    design_config.add_options()
      ("start-index,s",
       po::value<int>(&o.start_index)->default_value(1),
       "first index in FASTA file of sequences to optimize")
      ("end-index,e",
       po::value<int>(&o.end_index)->default_value(1),
       "last index in FASTA file of sequences to optimize")
      ("forbidden-motif-file,f", po::value< string > (&o.forbidden_filename),
       "a newline separated file containing forbidden motifs")
      ("desired-motif-file,d", po::value< string > (&o.desired_filename),
       "a newline separated file containing desired motifs")
      ;

    // Declare a group of options for outputting the
    // results of the analysis
    po::options_description other("Other");
    other.add_options()
      ("optimized-sequence-file,o",
       po::value<string>(&o.opt_seq_filename)->default_value("optimized.fasta"),
       "Output file for optimized sequences.")
      ("trace-file,t",
       po::value<string>(&o.trace_filename)->default_value("optimized.trace"),
       "Trace file for optimized sequences.")
      ;

    // Hidden options, will be allowed both on command line and
    // in config file, but will not be shown to the user.
    po::options_description hidden("Hidden options");
    hidden.add_options()
      ("fasta-file",
       po::value<string>(&o.fasta_file),
       "input file of oligo design")
      ;

    po::options_description cmdline_options;
    cmdline_options.add(generic).add(design_config).add(other).add(hidden);

    po::options_description config_file_options;
    config_file_options.add(design_config).add(other).add(hidden);

    po::options_description visible("Allowed options");
    visible.add(generic).add(design_config).add(other);

    po::positional_options_description p;
    p.add("fasta-file", 1);

    store(po::command_line_parser(argc, argv).
	  options(cmdline_options).positional(p).run(), vm);

    // if a configuration file is in current directory, parse options
    std::ifstream ifs("codon-optimizer.cfg");
    store(parse_config_file(ifs, config_file_options), vm);
    notify(vm);

    // check if the positional argument is correct, or if help was requested
    if (vm.count("help") || !vm.count("fasta-file")) {
      cout << endl
           << "Usage: codon-optimizer [options] <fasta_file>"
           << endl;
      cout << endl << visible << endl;
      return false;
    }

    if (o.forbidden_filename != "") {
      string tmp;
      ifstream file_forbidden(o.forbidden_filename.c_str());
      if (file_forbidden.is_open()) {
	while (! file_forbidden.eof() ) {
	  getline (file_forbidden, tmp);
	  if(tmp.length() >= 1) {
	    o.forbidden.push_back(DNASequence(tmp));
	  }
	}
	file_forbidden.close();
      }  else {
	cerr << "Unable to open file: " << o.forbidden_filename << ". Motifs in this file will be ignored.\n";
      }
    }

    if (o.desired_filename != "") {
      string tmp;
      ifstream file_desired(o.desired_filename.c_str());
      if (file_desired.is_open()) {
	while (! file_desired.eof() ) {
	  getline (file_desired, tmp);
	  if(tmp.length() >= 1) {
	    o.desired.push_back(DNASequence(tmp));
	  }
	}
	file_desired.close();
      } else {
	cerr << "Unable to open file: " << o.desired_filename << ". Motifs in this file will be ignored.\n";
      }
    } 
    
  }
  catch (exception& e) {
    cout << e.what() << endl;
    return false;
  }

  return true;
}


